# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
import time

import mock

from .. import worker

if sys.version[0] == '2':
    import Queue
else:
    import queue as Queue


ADDR = 'inproc://pour_the_rice_in_the_thing'
SETUP_SUCCESS_RETVAL = 'job setup success'


def test_run_with_timeout():
    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    msgid = worker._run_job(payload, logging.getLogger())

    assert msgid


@mock.patch('eventmq.worker.callable_from_name')
def test_run_job_setup_hook(callable_from_name_mock):
    from eventmq import conf

    setup_func_str = 'eventmq.tests.test_worker.job_setup_hook'
    callable_from_name_mock.return_value = mock.Mock()

    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    q, res_q = Queue.Queue(), Queue.Queue()

    q.put(payload)
    q.put('DONE')

    try:
        conf.JOB_ENTRY_FUNC = setup_func_str
        worker._run(q, res_q, logging.getLogger())
    finally:
        conf.JOB_ENTRY_FUNC = ''

    callable_from_name_mock.assert_called_with(setup_func_str)
    assert callable_from_name_mock.return_value.call_count == 1


@mock.patch('eventmq.worker.callable_from_name')
def test_run_job_teardown_hook(callable_from_name_mock):
    from eventmq import conf

    teardown_func_str = 'eventmq.tests.test_worker.job_teardown_hook'
    callable_from_name_mock.return_value = mock.Mock()

    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    q, res_q = Queue.Queue(), Queue.Queue()

    q.put(payload)
    q.put('DONE')

    try:
        conf.JOB_EXIT_FUNC = teardown_func_str
        worker._run(q, res_q, logging.getLogger())
    finally:
        conf.JOB_EXIT_FUNC = ''

    callable_from_name_mock.assert_called_with(teardown_func_str)
    assert callable_from_name_mock.return_value.call_count == 1


@mock.patch('eventmq.worker.callable_from_name')
def test_run_subprocess_setup_func(callable_from_name_mock):
    from eventmq import conf

    setup_func_str = 'eventmq.tests.test_worker.process_setup_hook'
    callable_from_name_mock.return_value = mock.Mock()

    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    q, res_q = Queue.Queue(), Queue.Queue()

    q.put(payload)
    q.put('DONE')

    try:
        conf.SUBPROCESS_SETUP_FUNC = setup_func_str
        worker._run(q, res_q, logging.getLogger())
    finally:
        conf.SUBPROCESS_SETUP_FUNC = ''

    callable_from_name_mock.assert_called_with(setup_func_str)
    assert callable_from_name_mock.return_value.call_count == 1


@mock.patch('eventmq.worker.run_setup')
def test_run_run_setup_func(run_setup_mock):
    from eventmq import conf

    setup_func_path = 'eventmq.tests.test_worker'
    setup_func_callable = 'process_setup_hook'

    payload = {
        'path': 'eventmq.tests.test_worker',
        'callable': 'job',
        'args': [2]
    }

    q, res_q = Queue.Queue(), Queue.Queue()

    q.put(payload)
    q.put('DONE')

    try:
        conf.SETUP_PATH = setup_func_path
        conf.SETUP_CALLABLE = setup_func_callable
        worker._run(q, res_q, logging.getLogger())
    finally:
        conf.SETUP_PATH = ''
        conf.SETUP_CALLABLE = ''

    run_setup_mock.assert_called_with(setup_func_path, setup_func_callable)


def test_run_setup():
    setup_callable = 'process_setup_hook'
    setup_path = 'eventmq.tests.test_worker'

    assert worker.run_setup(setup_path, setup_callable)


def job(sleep_time=0):
    time.sleep(sleep_time)

    return True


def process_setup_hook():
    return True


def job_setup_hook():
    return SETUP_SUCCESS_RETVAL


def job_teardown_hook():
    return True



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`publisher` -- Publisher
=======================
Publishes messages to subscribers
"""
import logging

import zmq

from . import constants
from .utils.devices import generate_device_name

logger = logging.getLogger(__name__)


class Publisher():
    """
        name (str): Name of this socket
        zcontext (:class:`zmq.Context`): socket context
        zsocket (:class:`zmq.Socket`):
    """

    def __init__(self, *args, **kwargs):
        self.zcontext = kwargs.get('context', zmq.Context.instance())
        self.name = kwargs.get('name', generate_device_name())

        self.zsocket = kwargs.get('socket', self.zcontext.socket(zmq.PUB))
        self.zsocket.setsockopt(zmq.IDENTITY, self.name)

        self.status = constants.STATUS.ready

        return

    def listen(self, addr=None):
        """
        listen to address defined by `addr`

        Args:
            addr (str): Address to listen to as a connection string

        Raises:
            :class:`Exception`
        """
        if self.ready:
            self.zsocket.bind(addr)
            self.status = constants.STATUS.connected
            logger.debug('Publisher %s: Listen to %s' % (self.name, addr))
        else:
            raise Exception('Receiver %s not ready. status=%s' %
                            (self.name, self.status))

    def publish(self, topic, msg):
        logger.debug("Notifying topic: {}".format(topic))
        return self.zsocket.send_multipart([topic, msg])

    @property
    def ready(self):
        """
        Property used to check if this receiver is ready.

        Returns:
            bool: True if the receiver is ready to connect or listen, otherwise
                False
        """
        return self.status == constants.STATUS.ready



# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
"""
:mod:`settings` -- Settings Utilities
=====================================
"""
from configparser import ConfigParser, NoOptionError
import json
import logging
import os

from . import tuplify
from .. import conf


logger = logging.getLogger(__name__)


def import_settings(section='global'):
    """
    Import settings and apply to configuration globals. This function will
    also read from the environment variables and override any value defined
    in the file.

    Args:
       section (str): Name of the INI section to import
    """
    config = ConfigParser()

    config_path = os.environ.get('EVENTMQ_CONFIG_FILE', conf.CONFIG_FILE)
    use_config_file = False

    if os.path.exists(config_path):
        config.read(config_path)
        if config.has_section(section):
            use_config_file = True
        else:
            logger.warning(
                'Tried to read nonexistent section {} from {}'.format(
                    section, config_path))

        for name in dir(conf):
            if name.startswith('_'):
                continue

            value = None
            found_value = False
            default_value = getattr(conf, name)

            # Favor environment variables over the config file definition
            try:
                value = os.environ['EVENTMQ_{}'.format(name)]
                found_value = True
            except KeyError:
                if use_config_file:
                    try:
                        value = config.get(section, name)
                        found_value = True
                    except NoOptionError:
                        found_value = False

            if found_value:
                t = type(getattr(conf, name))

                if t in (list, tuple):
                    try:
                        value = t(json.loads(value))
                    except ValueError:
                        raise ValueError(
                            "Invalid JSON syntax for {} setting".format(name))
                    # json.loads coverts all arrays to lists, but if the first
                    # element in the default is a tuple (like in QUEUES) then
                    # convert those elements, otherwise whatever it's type is
                    # correct
                    if isinstance(default_value[0], tuple):
                        setattr(conf, name, t(map(tuplify, value)))
                    else:
                        setattr(conf, name, t(value))
                elif isinstance(default_value, bool):
                    setattr(conf, name,
                            True if 't' in value.lower() else False)
                else:
                    setattr(conf, name, t(value))

            logger.debug("Setting conf.{} to {}".format(
                name, getattr(conf, name)))



# -*- coding: utf-8 -*-
# This file is part of eventmq.
#
# eventmq is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# eventmq is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with eventmq.  If not, see <http://www.gnu.org/licenses/>.
import json
import unittest

from .. import constants, scheduler

ADDR = 'inproc://pour_the_rice_in_the_thing'


class TestCase(unittest.TestCase):
    def test__setup(self):
        sched = scheduler.Scheduler(name='RuckusBringer')
        self.assertEqual(sched.name, 'RuckusBringer')

        self.assertFalse(sched.awaiting_startup_ack)
        self.assertEqual(sched.status, constants.STATUS.ready)

    def test_schedule_hash(self):
        msg1 = [
            'default',
            '',
            '3',
            json.dumps(['run', {
                'path': 'test',
                'args': [33, 'asdf'],
                'kwargs': {'zeta': 'Z', 'alpha': 'α'},
                'class_args': [0],
                'class_kwargs': {
                    'donkey': True, 'apple': False},
                'callable': 'do_the_thing'}]),
            None
        ]
        h1 = scheduler.Scheduler.schedule_hash(msg1)
        self.assertEqual('4658982cab9d32bf1ef9113a9d8bdec01775e2bc', h1)

        # Reordering the message argument shouldn't change the hash value
        msg2 = [
            'default',
            '',
            '3',
            json.dumps(['run', {
                'class_kwargs': {
                    'apple': False, 'donkey': True},
                'args': [33, 'asdf'],
                'class_args': [0],
                'kwargs': {'alpha': 'α', 'zeta': 'Z'},
                'path': 'test',
                'callable': 'do_the_thing'}]),
            None
        ]
        h2 = scheduler.Scheduler.schedule_hash(msg2)
        self.assertEqual('4658982cab9d32bf1ef9113a9d8bdec01775e2bc', h2)

# EMQP Tests
    def test_reset(self):
        sched = scheduler.Scheduler()

        self.assertFalse(sched.awaiting_startup_ack)
        self.assertEqual(sched.status, constants.STATUS.ready)
